/********************************************************************/
/*                                                                  */
/*  s7   Seed7 interpreter                                          */
/*  Copyright (C) 1990 - 2008  Thomas Mertes                        */
/*                                                                  */
/*  This program is free software; you can redistribute it and/or   */
/*  modify it under the terms of the GNU General Public License as  */
/*  published by the Free Software Foundation; either version 2 of  */
/*  the License, or (at your option) any later version.             */
/*                                                                  */
/*  This program is distributed in the hope that it will be useful, */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of  */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the   */
/*  GNU General Public License for more details.                    */
/*                                                                  */
/*  You should have received a copy of the GNU General Public       */
/*  License along with this program; if not, write to the           */
/*  Free Software Foundation, Inc., 51 Franklin Street,             */
/*  Fifth Floor, Boston, MA  02110-1301, USA.                       */
/*                                                                  */
/*  Module: Library                                                 */
/*  File: seed7/src/typlib.c                                        */
/*  Changes: 1993, 1994, 1999, 2000  Thomas Mertes                  */
/*  Content: All primitive actions for the type type.               */
/*                                                                  */
/********************************************************************/

#include "version.h"

#include "stdlib.h"
#include "stdio.h"
#include "string.h"

#include "common.h"
#include "data.h"
#include "data_rtl.h"
#include "heaputl.h"
#include "flistutl.h"
#include "datautl.h"
#include "syvarutl.h"
#include "striutl.h"
#include "object.h"
#include "typeutl.h"
#include "executl.h"
#include "objutl.h"
#include "runerr.h"
#include "traceutl.h"
#include "typ_data.h"

#undef EXTERN
#define EXTERN
#include "typlib.h"


#define TYPE_TABLE_INCREMENT 127



objectType typ_addinterface (listType arguments)

  {
    typeType typ1;
    typeType typ2;

  /* typ_addinterface */
    isit_type(arg_1(arguments));
    isit_type(arg_2(arguments));
    typ1 = take_type(arg_1(arguments));
    typ2 = take_type(arg_2(arguments));
    add_interface(typ1, typ2);
    return SYS_EMPTY_OBJECT;
  } /* typ_addinterface */



objectType typ_cmp (listType arguments)

  {
    memSizeType typ1;
    memSizeType typ2;
    intType signumValue;

  /* typ_cmp */
    isit_type(arg_1(arguments));
    isit_type(arg_2(arguments));
    typ1 = (memSizeType) take_type(arg_1(arguments));
    typ2 = (memSizeType) take_type(arg_2(arguments));
    if (typ1 < typ2) {
      signumValue = -1;
    } else if (typ1 > typ2) {
      signumValue = 1;
    } else {
      signumValue = 0;
    } /* if */
    return bld_int_temp(signumValue);
  } /* typ_cmp */



objectType typ_cpy (listType arguments)

  {
    objectType type_variable;

  /* typ_cpy */
    type_variable = arg_1(arguments);
    isit_type(type_variable);
    is_variable(type_variable);
    isit_type(arg_3(arguments));
    type_variable->value.typeValue = take_type(arg_3(arguments));
    return SYS_EMPTY_OBJECT;
  } /* typ_cpy */



objectType typ_create (listType arguments)

  {
    objectType type_to;
    typeType type_from;

  /* typ_create */
#ifdef TRACE_TYPLIB
    printf("BEGIN typ_create\n");
#endif
    isit_type(arg_3(arguments));
    type_to = arg_1(arguments);
    type_from = take_type(arg_3(arguments));
#ifdef TRACE_typ_create
    printf("\nbefore type assignment\n");
    trace1(type_to);
    printf("\n");
    trace1(type_from->match_obj);
    printf("\n");
#endif
    SET_CATEGORY_OF_OBJ(type_to, TYPEOBJECT);
    type_to->value.typeValue = type_from;
    if (!VAR_OBJECT(type_to)) {
      if (type_from->name == NULL &&
          HAS_ENTITY(type_to) &&
          GET_ENTITY(type_to)->ident != NULL) {
        /* printf("typ_create: Define ");
        trace1(type_to);
        printf("\n"); */
        type_from->name = GET_ENTITY(type_to)->ident;
      } /* if */
    } /* if */
#ifdef TRACE_TYPLIB
    printf("END typ_create\n");
#endif
    return SYS_EMPTY_OBJECT;
  } /* typ_create */



objectType typ_destr (listType arguments)

  {
    /* typeType old_type; */

  /* typ_destr */
#ifdef TRACE_TYPLIB
    printf("BEGIN typ_destr\n");
#endif
    isit_type(arg_1(arguments));
    /* old_type = take_type(arg_1(arguments)); */
    /* printf("typ_destr "); trace1(old_type->match_obj); printf("\n"); */
    SET_UNUSED_FLAG(arg_1(arguments));
#ifdef TRACE_TYPLIB
    printf("END typ_destr\n");
#endif
    return SYS_EMPTY_OBJECT;
  } /* typ_destr */



objectType typ_eq (listType arguments)

  {
    typeType type1;
    typeType type2;
    objectType result;

  /* typ_eq */
    isit_type(arg_1(arguments));
    isit_type(arg_3(arguments));
    type1 = take_type(arg_1(arguments));
    type2 = take_type(arg_3(arguments));
    if (type1 == type2) {
      result = SYS_TRUE_OBJECT;
    } else {
      result = SYS_FALSE_OBJECT;
    } /* if */
    return result;
  } /* typ_eq */



objectType typ_func (listType arguments)

  {
    typeType basic_type;
    typeType result;

  /* typ_func */
#ifdef TRACE_TYPLIB
    printf("BEGIN typ_func\n");
#endif
    isit_type(arg_2(arguments));
    basic_type = take_type(arg_2(arguments));
    if ((result = get_func_type(NULL, basic_type)) == NULL) {
      return raise_exception(SYS_MEM_EXCEPTION);
    } /* if */
    /* printf("typ_func ");
    printobject(result);
    prot_cstri("=");
    prot_int((intType) result);
    printf("\n"); */
#ifdef TRACE_TYPLIB
    printf("END typ_func\n");
#endif
    return bld_type_temp(result);
  } /* typ_func */



objectType typ_gensub (listType arguments)

  {
    typeType meta_type;
    typeType result;

  /* typ_gensub */
#ifdef TRACE_TYPLIB
    printf("BEGIN typ_gensub\n");
#endif
    isit_type(arg_1(arguments));
    meta_type = take_type(arg_1(arguments));
    if ((result = new_type(meta_type->owningProg, meta_type, NULL)) == NULL) {
      return raise_exception(SYS_MEM_EXCEPTION);
    } /* if */
#ifdef TRACE_TYPLIB
    printf("END typ_gensub\n");
#endif
    return bld_type_temp(result);
  } /* typ_gensub */



objectType typ_gentype (listType arguments)

  {
    typeType result;

  /* typ_gentype */
#ifdef TRACE_TYPLIB
    printf("BEGIN typ_gentype\n");
#endif
    if ((result = new_type(prog.owningProg, NULL, NULL)) == NULL) {
      return raise_exception(SYS_MEM_EXCEPTION);
    } /* if */
#ifdef TRACE_TYPLIB
    printf("END typ_gentype\n");
#endif
    return bld_type_temp(result);
  } /* typ_gentype */



#ifdef OUT_OF_ORDER
objectType typ_getinterfaces (listType arguments)

  {
    typeType typ1;

  /* typ_getinterfaces */
    isit_type(arg_1(arguments));
    typ1 = take_type(arg_1(arguments));
    get_interfaces(typ1);
    return SYS_EMPTY_OBJECT;
  } /* typ_getinterfaces */
#endif



#ifdef OUT_OF_ORDER
objectType typ_getcreate (listType arguments)

  {
    typeType result;

  /* typ_getcreate */
    isit_type(arg_1(arguments));
    ;
    get_create_call_obj(take_type(arg_1(arguments)), errInfoType *err_info)
    if ((result = new_type(prog.owningProg, NULL, NULL)) == NULL) {
      return raise_exception(SYS_MEM_EXCEPTION);
    } /* if */
    return bld_type_temp(result);
  } /* typ_getcreate */
#endif



objectType typ_hashcode (listType arguments)

  { /* typ_hashcode */
    isit_type(arg_1(arguments));
    return bld_int_temp((intType)
        (((memSizeType) take_type(arg_1(arguments))) >> 6));
  } /* typ_hashcode */



objectType typ_isdeclared (listType arguments)

  {
    objectType result;

  /* typ_isdeclared */
    if (CATEGORY_OF_OBJ(arg_1(arguments)) == DECLAREDOBJECT) {
      result = SYS_TRUE_OBJECT;
    } else {
      result = SYS_FALSE_OBJECT;
    } /* if */
    return result;
  } /* typ_isdeclared */



objectType typ_isderived (listType arguments)

  {
    typeType any_type;
    objectType result;

  /* typ_isderived */
    isit_type(arg_1(arguments));
    any_type = take_type(arg_1(arguments));
    if (any_type->meta != NULL) {
      result = SYS_TRUE_OBJECT;
    } else {
      result = SYS_FALSE_OBJECT;
    } /* if */
    return result;
  } /* typ_isderived */



objectType typ_isforward (listType arguments)

  {
    objectType result;

  /* typ_isforward */
    if (CATEGORY_OF_OBJ(arg_1(arguments)) == FORWARDOBJECT) {
      result = SYS_TRUE_OBJECT;
    } else {
      result = SYS_FALSE_OBJECT;
    } /* if */
    return result;
  } /* typ_isforward */



objectType typ_isfunc (listType arguments)

  {
    typeType any_type;
    objectType result;

  /* typ_isfunc */
    isit_type(arg_1(arguments));
    any_type = take_type(arg_1(arguments));
    if (any_type->result_type != NULL && !any_type->is_varfunc_type) {
      result = SYS_TRUE_OBJECT;
    } else {
      result = SYS_FALSE_OBJECT;
    } /* if */
    return result;
  } /* typ_isfunc */



objectType typ_isvarfunc (listType arguments)

  {
    typeType any_type;
    objectType result;

  /* typ_isvarfunc */
    isit_type(arg_1(arguments));
    any_type = take_type(arg_1(arguments));
    if (any_type->result_type != NULL && any_type->is_varfunc_type) {
      result = SYS_TRUE_OBJECT;
    } else {
      result = SYS_FALSE_OBJECT;
    } /* if */
    return result;
  } /* typ_isvarfunc */



objectType typ_matchobj (listType arguments)

  {
    typeType actual_type;
    objectType result;

  /* typ_matchobj */
    isit_type(arg_1(arguments));
    actual_type = take_type(arg_1(arguments));
    result = actual_type->match_obj;
    return bld_reference_temp(result);
  } /* typ_matchobj */



objectType typ_meta (listType arguments)

  {
    typeType any_type;
    typeType result;

  /* typ_meta */
#ifdef TRACE_TYPLIB
    printf("BEGIN typ_meta\n");
#endif
    isit_type(arg_1(arguments));
    any_type = take_type(arg_1(arguments));
    if (any_type->meta == NULL) {
      return raise_exception(SYS_RNG_EXCEPTION);
    } /* if */
    result = any_type->meta;
#ifdef TRACE_TYPLIB
    printf("END typ_meta\n");
#endif
    return bld_type_temp(result);
  } /* typ_meta */



objectType typ_ne (listType arguments)

  {
    typeType type1;
    typeType type2;
    objectType result;

  /* typ_ne */
    isit_type(arg_1(arguments));
    isit_type(arg_3(arguments));
    type1 = take_type(arg_1(arguments));
    type2 = take_type(arg_3(arguments));
    if (type1 != type2) {
      result = SYS_TRUE_OBJECT;
    } else {
      result = SYS_FALSE_OBJECT;
    } /* if */
    return result;
  } /* typ_ne */



objectType typ_num (listType arguments)

  { /* typ_num */
    isit_type(arg_1(arguments));
    return bld_int_temp(
        typNum(take_type(arg_1(arguments))));
  } /* typ_num */



objectType typ_result (listType arguments)

  {
    typeType any_type;
    typeType result;

  /* typ_result */
#ifdef TRACE_TYPLIB
    printf("BEGIN typ_result\n");
#endif
    isit_type(arg_1(arguments));
    any_type = take_type(arg_1(arguments));
    if (any_type->result_type == NULL) {
      /* trace1(arg_1(arguments));
      printf("\n"); */
      return raise_exception(SYS_RNG_EXCEPTION);
    } /* if */
    result = any_type->result_type;
#ifdef TRACE_TYPLIB
    printf("END typ_result\n");
#endif
    return bld_type_temp(result);
  } /* typ_result */



objectType typ_str (listType arguments)

  { /* typ_str */
    isit_type(arg_1(arguments));
    return bld_stri_temp(
        typStr(take_type(arg_1(arguments))));
  } /* typ_str */



objectType typ_value (listType arguments)

  {
    objectType obj_arg;

  /* typ_value */
    isit_reference(arg_1(arguments));
    obj_arg = take_reference(arg_1(arguments));
    if (obj_arg == NULL || CATEGORY_OF_OBJ(obj_arg) != TYPEOBJECT) {
      return raise_exception(SYS_RNG_EXCEPTION);
    } else {
      return obj_arg;
    } /* if */
  } /* typ_value */



objectType typ_varconv (listType arguments)

  { /* typ_varconv */
    is_variable(arg_3(arguments));
    if (TEMP_OBJECT(arg_3(arguments))) {
      prot_cstri("TYP_VARCONV of TEMP_OBJECT ");
      trace1(arg_3(arguments));
      prot_nl();
    } /* if */
    return arg_3(arguments);
  } /* typ_varconv */



objectType typ_varfunc (listType arguments)

  {
    typeType basic_type;
    typeType result;

  /* typ_varfunc */
#ifdef TRACE_TYPLIB
    printf("BEGIN typ_varfunc\n");
#endif
    isit_type(arg_2(arguments));
    basic_type = take_type(arg_2(arguments));
    if ((result = get_varfunc_type(NULL, basic_type)) == NULL) {
      return raise_exception(SYS_MEM_EXCEPTION);
    } /* if */
    /* printf("typ_varfunc ");
    printobject(result);
    prot_cstri("=");
    prot_int((intType) result);
    printf("\n"); */
#ifdef TRACE_TYPLIB
    printf("END typ_varfunc\n");
#endif
    return bld_type_temp(result);
  } /* typ_varfunc */
